import unittest
from nbt2yaml import parse_nbt, dump_nbt
from tests import datafile, eq_
import gzip
import StringIO

class DumpNBTTest(unittest.TestCase):
    def _assert_data(self, fname):
        unzipped_data = gzip.GzipFile(fileobj=datafile(fname)).read()
        parsed = parse_nbt(datafile(fname))
        out = StringIO.StringIO()
        dump_nbt(parsed, out)
        out = gzip.GzipFile(fileobj=StringIO.StringIO(out.getvalue())).read()
        eq_(unzipped_data, out)

    def test_basic(self):
        self._assert_data("test.nbt")

    def test_large(self):
        self._assert_data("bigtest.nbt")



import argparse
import sys
from nbt2yaml import parse_nbt, dump_yaml, parse_yaml, dump_nbt
import tempfile
import shutil
import StringIO
import os

def nbtedit():
    parser = argparse.ArgumentParser(description="Edit an nbt file in-place in yaml format.")
    parser.add_argument("filename", type=str, help="filename")
    parser.add_argument("-n", "--no-gzip", 
                        action="store_true", 
                        help="Don't use gzip"
                        )

    options = parser.parse_args()

    struct = parse_nbt(
                    open(options.filename, 'rb'), 
                    gzipped=not options.no_gzip)
    try:
        editor = os.environ['EDITOR']
    except KeyError:
        sys.exit("Environment variable EDITOR is not set")

    edit_file = tempfile.NamedTemporaryFile(delete=False, suffix='.yml')
    edit_file.write(dump_yaml(struct))
    edit_file.close()
    edit_filename = edit_file.name
    os.system("%s %s" % (editor, edit_filename))


    edit_file = open(edit_filename, 'rb')
    new_struct = parse_yaml(edit_file)
    edit_file.close()
    os.remove(edit_filename)

    if new_struct == struct:
        sys.stderr.write("No changes made\n")
        sys.exit()

    save_counter = 0
    while True:
        if save_counter == 0:
            savefile = os.extsep.join([options.filename, "saved"])
        else:
            savefile = os.extsep.join([options.filename, str(save_counter), "saved"])
        if not os.path.exists(savefile):
            break
        save_counter += 1

    shutil.move(options.filename, savefile)
    sys.stderr.write("Saving old file as %s\n" % savefile)
    write_file = open(options.filename, 'wb')
    sys.stderr.write("Writing %s\n" % options.filename)
    dump_nbt(new_struct, write_file, gzipped=not options.no_gzip)
    write_file.close()

def nbt2yaml():
    parser = argparse.ArgumentParser(description="Dump an nbt file or stream to yaml.")
    parser.add_argument("filename", type=str, help="Filename.  Specify as '-' to read from stdin.")
    parser.add_argument("-n", "--no-gzip", 
                        action="store_true", 
                        help="Don't use gzip"
                        )
    options = parser.parse_args()
    if options.filename == '-':
        input_ = StringIO.StringIO(sys.stdin.read())
    else:
        input_ = open(options.filename, 'rb')

    struct = parse_nbt(input_, gzipped=not options.no_gzip)
    print dump_yaml(struct)

def yaml2nbt():
    parser = argparse.ArgumentParser(description="Dump a yaml file or stream to nbt.")
    parser.add_argument("filename", type=str, help="Filename.  Specify as '-' to read from stdin.")
    parser.add_argument("-n", "--no-gzip", 
                        action="store_true", 
                        help="Don't use gzip"
                        )
    options = parser.parse_args()
    if options.filename == '-':
        input_ = StringIO.StringIO(sys.stdin.read())
    else:
        input_ = open(options.filename, 'rb')

    struct = parse_yaml(input_)
    dump_nbt(struct, sys.stdout)



import unittest
from nbt2yaml import parse_nbt, parse
from tests import datafile, eq_
import struct

class ParseNBTTest(unittest.TestCase):
    def test_basic(self):
        assert_ = (
                    parse.TAG_Compound, 
                    u'hello world', 
                    [(parse.TAG_String, u'name', u'Bananrama')]
                )
        eq_(parse_nbt(datafile("test.nbt")), assert_)

    def test_large(self):
        assert_ = (
                    parse.TAG_Compound, u'Level', [
                        (parse.TAG_Long, u'longTest', 9223372036854775807L), 
                        (parse.TAG_Short, u'shortTest', 32767), 
                        (parse.TAG_String, u'stringTest', u'HELLO WORLD THIS IS A TEST STRING \xc5\xc4\xd6!'), 
                        (parse.TAG_Float, u'floatTest', 0.4982314705848694), 
                        (parse.TAG_Int, u'intTest', 2147483647), 
                        (parse.TAG_Compound, u'nested compound test', 
                            [
                                (parse.TAG_Compound, u'ham', 
                                    [
                                        (parse.TAG_String, u'name', u'Hampus'), 
                                        (parse.TAG_Float, u'value', 0.75)
                                    ]), 
                                    (parse.TAG_Compound, u'egg', 
                                    [
                                        (parse.TAG_String, u'name', u'Eggbert'), 
                                        (parse.TAG_Float, u'value', 0.5)
                                    ])
                                ]), 
                                (parse.TAG_List, u'listTest (long)', (parse.TAG_Long, [11, 12, 13, 14, 15])), 
                    (parse.TAG_List, u'listTest (compound)', 
                        (parse.TAG_Compound, [
                            [
                                (parse.TAG_String, u'name', u'Compound tag #0'), 
                                (parse.TAG_Long, u'created-on', 1264099775885L)
                            ], 
                            [
                                (parse.TAG_String, u'name', u'Compound tag #1'), 
                                (parse.TAG_Long, u'created-on', 1264099775885L)
                            ]
                        ])), 
                        (parse.TAG_Byte, u'byteTest', 127), 
                        (parse.TAG_Byte_Array, 
                            u'byteArrayTest (the first 1000 values of '
                                    '(n*n*255+n*7)%100, starting with n=0 '
                                    '(0, 62, 34, 16, 8, ...))', 
                                    ''.join([struct.pack('>b', (n*n*255+n*7)%100) for n in range(0, 1000)])
                            ),
                        (parse.TAG_Double, u'doubleTest', 0.4931287132182315)
                    ])
        eq_(parse_nbt(datafile("bigtest.nbt")), assert_)



import unittest
from nbt2yaml import parse_nbt, dump_yaml, parse_yaml, dump_nbt
from tests import datafile, eq_

class FromYamlTest(unittest.TestCase):
    def test_basic(self):
        data = parse_yaml(datafile("test.yml"))

        eq_(data, parse_nbt(datafile("test.nbt")))

    def test_large(self):
        data = parse_yaml(datafile("bigtest.yml"))

        eq_(data, parse_nbt(datafile("bigtest.nbt")))

    def test_lists(self):
        data = parse_yaml(datafile("list.yml"))

        eq_(data, parse_nbt(datafile("list.nbt")))


import struct
import gzip
from collections import namedtuple
import re

class Tag(object):
    _tags = {}
    _tuple = namedtuple("Tag", ["type", "name", "data"])

    def __init__(self, name, id):
        self.name = name
        self.id = id
        Tag._tags[id] = self

    @classmethod
    def from_stream(cls, stream):
        return cls._tags[struct.unpack('>b', stream.read(1))[0]]

    def to_stream(self, stream):
        stream.write(struct.pack('>b', self.id))

    def parse(self, stream):
        name = TAG_String._parse_impl(stream)
        data = self._parse_impl(stream)
        return Tag._tuple(self, name, data)

    def dump(self, element, stream):
        type_, name, data = element
        type_.to_stream(stream)
        TAG_String._dump_impl(name, stream)
        self._dump_impl(data, stream)

    def _parse_length(self, length_type, stream):
        return struct.unpack(">" + length_type.format, 
                        stream.read(length_type.size))[0]

    def _dump_length(self, length_type, length, stream):
        stream.write(
            struct.pack(">" + length_type.format, length)
        )

    def _parse_impl(self, stream):
        raise NotImplementedError()

    def _dump_impl(self, data, stream):
        raise NotImplementedError()

    def __repr__(self):
        return "TAG_%s" % \
            re.sub(r'^(\w)|_(\w)', 
                    lambda m:m.group(0).upper(), self.name)

class FixedTag(Tag):
    def __init__(self, name, id, size, format):
        Tag.__init__(self, name, id)
        self.size = size
        self.format = format

    def _parse_impl(self, stream):
        return struct.unpack(">" + self.format, stream.read(self.size))[0]

    def _dump_impl(self, data, stream):
        stream.write(
            struct.pack(">" + self.format, data)
        )

class EndTag(Tag):
    ""

class VariableTag(Tag):
    def __init__(self, name, id, length_tag, encoding=None):
        Tag.__init__(self, name, id)
        self.length_tag = length_tag
        self.encoding = encoding

    def _parse_impl(self, stream):
        length = self._parse_length(self.length_tag, stream)
        data = stream.read(length)
        if self.encoding:
            data = data.decode(self.encoding)
        return data

    def _dump_impl(self, data, stream):
        if self.encoding:
            data = data.encode(self.encoding)
        self._dump_length(self.length_tag, len(data), stream)
        stream.write(data)

class ListTag(Tag):
    def _parse_impl(self, stream):
        element_type = Tag.from_stream(stream)
        length = self._parse_length(TAG_Int, stream)

        return (element_type, [
            element_type._parse_impl(stream) for i in xrange(length)
        ])

    def _dump_impl(self, data, stream):
        element_type, data = data
        element_type.to_stream(stream)
        self._dump_length(TAG_Int, len(data), stream)
        for elem in data:
            element_type._dump_impl(elem, stream)

class CompoundTag(Tag):
    def _parse_impl(self, stream):
        data = []
        while True:
            c_type_ = Tag.from_stream(stream)
            if c_type_ is TAG_End:
                break
            data.append(c_type_.parse(stream))
        return data

    def _dump_impl(self, data, stream):
        for elem in data:
            elem[0].dump(elem, stream)
        TAG_End.to_stream(stream)

TAG_End = EndTag('end', 0)
TAG_Byte = FixedTag('byte', 1, 1, 'b')
TAG_Short = FixedTag('short', 2, 2, 'h')
TAG_Int = FixedTag('int', 3, 4, 'i')
TAG_Long = FixedTag('long', 4, 8, 'q')
TAG_Float = FixedTag('float', 5, 4, 'f')
TAG_Double = FixedTag('double', 6, 8, 'd')
TAG_Byte_Array = VariableTag('byte_array', 7, TAG_Int)
TAG_String = VariableTag('string', 8, TAG_Short, encoding='utf-8')
TAG_List = ListTag('list', 9)
TAG_Compound = CompoundTag('compound', 10)

def parse_nbt(stream, gzipped=True):
    """Parse an .nbt file from the given stream.
    
    :param stream: a file-like stream configured for read().
    :param gzipped: if True, assume the stream is in gzip format first.
    :return: a Python structure representing the .nbt file.

    """
    if gzipped:
        stream = gzip.GzipFile(fileobj=stream)
    type_ = Tag.from_stream(stream)
    return type_.parse(stream)

def dump_nbt(nbt, stream, gzipped=True):
    """Dump an .nbt structure generated by :func:`.parse_nbt` 
    to the given stream.
    
    :param nbt: the structure returned by a :func:`.parse_nbt` call.
    :param stream: a file-like stream configured for write().
    :param gzipped: if True, pass the stream through gzip for final output.
    
    """
    if gzipped:
        stream = gzip.GzipFile(mode='w', fileobj=stream)
    type_ = nbt[0]
    type_.dump(nbt, stream)



import os
import sys

from setuptools import setup, find_packages

extra = {}
if sys.version_info >= (3, 0):
    extra.update(
        use_2to3=True,
    )

readme = os.path.join(os.path.dirname(__file__), 'README.rst')

setup(name='nbt2yaml',
      version="0.2.0",
      description="Read and write Minecraft NBT files using Yaml.",
      long_description=file(readme).read(),
      classifiers=[
      'Development Status :: 3 - Alpha',
      'License :: OSI Approved :: BSD License',
      'Programming Language :: Python',
      'Programming Language :: Python :: 3',
      ],
      keywords='minecraft',
      author='Mike Bayer',
      author_email='mike_mp@zzzcomputing.com',
      url='http://bitbucket.org/zzzeek/nbt2yaml',
      license='BSD',
      packages=find_packages(exclude=['ez_setup', 'tests']),
      zip_safe=False,
      install_requires=['PyYAML'],
      scripts=[
                'scripts/nbtedit',
                'scripts/nbt2yaml',
                'scripts/yaml2nbt'
                ],
      test_suite='nose.collector',
      tests_require=['nose'],
      **extra
)



import yaml
from nbt2yaml import parse

_explicit_types = parse.TAG_Short, parse.TAG_Long, parse.TAG_Double, parse.TAG_Byte, parse.TAG_Byte_Array
_all_types = parse.Tag._tags.values()

_canned_types = {
    str:parse.TAG_String,
    float:parse.TAG_Float,
    int:parse.TAG_Int,
}

_types_to_python = {
    parse.TAG_Long:long,
    parse.TAG_Int:int,
    parse.TAG_Short:int,
    parse.TAG_Byte:int,
    parse.TAG_Float:float,
    parse.TAG_Double:float
}

class ForceType(object):
    """Represent a data value with an explicit type.
    
    This is used to output 'short', 'long', 'double', 'byte'
    explicitly, so that we can differentiate on the
    yaml parsing side what specific NBT form to use.
    
    """
    def __init__(self, type_, value):
        self.type = type_
        self.value = value

class ForceListOfType(ForceType):
    pass

def _type_representer(dumper, struct):
    if struct.type is parse.TAG_Byte_Array:
        representation = struct.value
    else:
        representation = repr(struct.value)
    return dumper.represent_scalar(u'!%s' % struct.type.name, representation, style='""')

def _collection_representer(dumper, struct):
    return dumper.represent_sequence(u'!list_%s' % struct.type.name, struct.value)

yaml.add_representer(ForceType, _type_representer)
yaml.add_representer(ForceListOfType, _collection_representer)

def _type_constructor(type_):
    def _constructor(loader, node):
        value = loader.construct_scalar(node)
        return ForceType(type_, value)
    return _constructor

def _list_constructor(type_):
    def _constructor(loader, node):
        value = loader.construct_sequence(node)
        return ForceListOfType(type_, value)
    return _constructor

for type_ in _explicit_types:
    yaml.add_constructor(u'!%s' % type_.name, _type_constructor(type_))

for type_ in _all_types:
    yaml.add_constructor(u'!list_%s' % type_.name, _list_constructor(type_))

def _yaml_serialize(struct):
    tag, name, data = struct.type, struct.name, struct.data
    name = name.encode('utf-8')
    return {name:_value_as_yaml(tag, data)}

def _yaml_deserialize(struct):
    name, data = struct.items()[0]
    type_ = _type_from_yaml(data)
    return parse.Tag._tuple(type_, name.decode('utf-8'), _yaml_as_value(type_, data))

def _value_as_yaml(type_, value):
    if type_ is parse.TAG_Compound:
        return [_yaml_serialize(s) for s in value]
    elif type_ is parse.TAG_String:
        return value.encode('utf-8')
    elif type_ in _explicit_types:
        return ForceType(type_, value)
    elif type_ is parse.TAG_List:
        element_type, data = value
        return ForceListOfType(element_type, [_value_as_yaml(element_type, d) for d in data])
    else:
        return value

def _yaml_as_value(type_, value):
    if type_ is parse.TAG_Compound:
        return [_yaml_deserialize(s) for s in value]
    elif type_ is parse.TAG_String:
        return value.decode('utf-8')
    elif type_ in _explicit_types:
        if type_ in _types_to_python:
            return _types_to_python[type_](value.value)
        else:
            return value.value
    elif type_ is parse.TAG_List:
        ltype = value.type
        return (ltype, [_yaml_as_value(ltype, s) for s in value.value])
    else:
        return value

def _type_from_yaml(data):
    if isinstance(data, ForceListOfType):
        return parse.TAG_List
    elif isinstance(data, list):
        assert data and isinstance(data[0], dict)
        return parse.TAG_Compound
    elif isinstance(data, ForceType):
        return data.type
    elif type(data) in _canned_types:
        return _canned_types[type(data)]
    else:
        raise ValueError("Can't determine type for element: %r" % data)

def dump_yaml(struct, canonical=False, default_flow_style=False):
    """Dump a yaml string given an nbt structure parsed by nbt2yaml."""

    return yaml.dump(
                _yaml_serialize(struct), 
                default_flow_style=default_flow_style, 
                canonical=canonical)


def parse_yaml(stream):
    """Parse a yaml stream into an nbt structure readable by nbt2yaml."""

    return _yaml_deserialize(yaml.load(stream))
